/* (c) 1996-2001, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* Tagfile Identification Database Interface
 *
 * This module supports using tagfiles as an identification database.  This
 * is a file named 'id' that sits in the user's home directory and has lines
 * like 'fullname=(Jan Wolter)'.  The directory path isn't stored in the
 * tagfile, because you need it before you can access the tagfile.  Instead
 * we generate the directory name from the login name.
 */


#include "backtalk.h"
#include "authident.h"

#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif

#include "set_uid.h"
#include "str.h"
#include "udb.h"
#include "lock.h"
#include "entropy.h"
#include "tag.h"

#ifdef DMALLOC
#include "dmalloc.h"
#endif

char *showopt_ident_module= "ident_tagfile";


/* SAVE_TAG - Write the given variable/value pairs into a tag file, replacing
 * any old values, if there were ones, or appending the new values.  The
 * values are not changed if they were protected and the user is not an
 * administrator.  Can set multiple values, like this:
 *    save_tag(fname, amadm, var1, type1, val1, NULL)
 *    save_tag(fname, amadm, var1, type1, val1, var2, type2, val2, NULL)
 */

void save_tag(char *fname, int amadm, ...)
{
    va_list ap;
    taglist *tagl= NULL;
    char *varname;
    int type;

    va_start(ap, amadm);
    while ((varname= va_arg(ap, char *)) != NULL)
    {
	/* Add our new tag to our tag list */
	type= va_arg(ap, int);
	if (type == TK_INTEGER)
	    add_int_tag(&tagl, varname, va_arg(ap, int));
	else
	    add_str_tag(&tagl, varname, va_arg(ap, char *));
    }

    update_tag(fname, &tagl, amadm);

    va_end(ap);
}


/* SVCPY - Copy the given string value into a buffer.  stp is our current
 * position in the buffer.  ste points at the last byte of the buffer.
 * copy's the string in, truncating it if it reachs the last byte.  Returns
 * a pointer either to the first character after the copied string, or ste
 * whichever is less.
 */

char *svcpy(char *val, char *stp, char *ste)
{
    while (*val != '\0' && stp < ste)
	*(stp++)= *(val++);
    *stp= '\0';
    return (stp < ste) ? stp+1 : ste;
}


/* IDENT_GETPWNAM - Get the database entry for the named user.
 */

#ifdef HAVEIDENT_GETPWNAM
struct btpasswd *ident_getpwnam(char *nam)
{
    static struct btpasswd pw;
    static char stbf[BFSZ];
    char fname[BFSZ+1];
    char *dir, *var, *val, *stp;
    rtag *rt;
    int ch, gotuid, gotgid;
    int rc;

    pw.name= nam;
    pw.email= pw.gecos= pw.valcode= NULL;
    pw.priv= pw.epriv= -1;
    gotuid= gotgid= FALSE;
    stp= stbf;

    /* Construct the directory name */
    if (nam == NULL || (dir= homedirpath(nam)) == NULL)
	return NULL;
    pw.dir= dir;

    /* Open tag file */
    sprintf(fname,"%s/"IDENT_FILE,dir);
    if ((rt= open_tag(fname,TRUE)) == NULL)
	    return NULL;
    
    /* Read through tag file */
    while ((rc= get_tag(rt, &var, &val)) > 0)
    {
	if (rc > 2) rc-= 2;

	if (!strcmp(var, "uid"))
	{
	    if (rc != 2) goto error;
	    pw.uid= atoi(val);
	    gotuid= TRUE;
	}
	else if (!strcmp(var, "gid"))
	{
	    if (rc != 2) goto error;
	    pw.gid= atoi(val);
	    gotgid= TRUE;
	}
	else if (!strcmp(var, "fullname"))
	{
	    if (rc != 1) goto error;
	    pw.gecos= stp;
	    stp= svcpy(val, stp, stbf+BFSZ);
	}
	else if (!strcmp(var, "email"))
	{
	    if (rc != 1) goto error;
	    pw.email= stp;
	    stp= svcpy(val, stp, stbf+BFSZ);
	}
	else if (!strcmp(var, "valcode"))
	{
	    if (rc != 1) goto error;
	    pw.valcode= stp;
	    stp= svcpy(val, stp, stbf+BFSZ);
	}
	else if (!strcmp(var, "privacy"))
	{
	    if (rc != 2) goto error;
	    pw.priv= atoi(val);
	}
	else if (!strcmp(var, "eprivacy"))
	{
	    if (rc != 2) goto error;
	    pw.epriv= atoi(val);
	}
    }

    if (gotuid && gotgid)
    {
	close_tag(rt);
	return &pw;
    }

error:
    close_tag(rt);
    return NULL;
}
#endif


/* IDENT_NEWFNAME - change the user's fullname to the given value.
 */

#ifdef HAVEIDENT_NEWFNAME
void ident_newfname(struct btpasswd *pw, char *newfname)
{
    char fname[BFSZ];

    /* Construct name of ident file */
    sprintf(fname,"%s/"IDENT_FILE,pw->dir);

    flushusercache();

    save_tag(fname, 0, "fullname", TK_STRING, newfname, NULL);
}
#endif


/* IDENT_NEWGID - change the user's groupid to the given value.
 */

#ifdef HAVEIDENT_NEWGID
void ident_newgid(struct btpasswd *pw, gid_t newgid)
{
    char fname[BFSZ];

    /* Construct name of ident file */
    sprintf(fname,"%s/"IDENT_FILE,pw->dir);

    flushusercache();

    save_tag(fname, 1, "gid", TK_INTEGER, newgid, NULL);
}
#endif

/* IDENT_NEWEMAIL - change the user's email address to the given value.
*  */

#ifdef HAVEIDENT_NEWEMAIL
void ident_newemail(struct btpasswd *pw, char *newemail, int newvalid)
{
    char fname[BFSZ];
    char *valcode= NULL;

    /* Construct name of ident file */
    sprintf(fname,"%s/"IDENT_FILE,pw->dir);

    flushusercache();

    if (newemail[0] != '\0' && !newvalid)
	valcode= get_noise();

    save_tag(fname, 1,
	    "email", TK_STRING, newemail,
	    "valcode", TK_STRING, valcode == NULL ? "" : valcode,
	    NULL);
}
#endif


#ifdef HAVEIDENT_NEWPRIV
void ident_newpriv(struct btpasswd *pw, int priv, int epriv)
{
    char fname[BFSZ];

    /* Construct name of ident file */
    sprintf(fname,"%s/"IDENT_FILE,pw->dir);

    flushusercache();

    save_tag(fname, 1,
	    "privacy", TK_INTEGER, priv,
	    "eprivacy", TK_INTEGER, epriv,
	    NULL);
}
#endif


/* IDENT_ADD - Add an entry to the user database describing a user with
 * the given login, uid, gid, password, fullname, and directory.
 */

#ifdef HAVEIDENT_ADD
void ident_add(char *lid, uid_t uid, gid_t gid, char *fn, char *dir,
	char *email, int valid, int priv, int epriv)
{
    char fname[BFSZ];
    char *valcode= NULL;
    FILE *fp;

    if (email != NULL && email[0] != '\0' && !valid)
	valcode= get_noise();

    sprintf(fname,"%s/"IDENT_FILE,dir);
    if ((fp= fopen(fname,"w")) == NULL)
	die("could not create ident file %s",fname);
    lock_exclusive(fileno(fp),fname);
    fprintf(fp,"*uid=%d\n*gid=%d\n", uid, gid);
    if (fn != NULL) wrtstrtag(fp,"fullname",fn,FALSE);
    if (email != NULL) wrtstrtag(fp,"email",email,FALSE);
    if (valcode != NULL) wrtstrtag(fp,"valcode",valcode,FALSE);
    if (priv != -1) wrtinttag(fp,"privacy", priv, FALSE);
    if (epriv != -1) wrtinttag(fp,"eprivacy", epriv, FALSE);
    fclose(fp);
    unlock(fname);
}
#endif


/* IDENT_DEL - Delete the named user from the ident database.  Actually, there
 * is nothing to do here, since we will always be deleting the directory
 * next, and that will take care of everything.
 */

#ifdef HAVEIDENT_DEL
void ident_del(char *name) { }
#endif
