/* Copyright 2005, Jan D. Wolter, All Rights Reserved. */

#include <fcntl.h>

#include "backtalk.h"
#include "sysdict.h"
#include "str.h"
#include "stack.h"
#include "subscriber.h"

/* SUBSCRIBER FILE ROUTINES
 *
 * In each conference directory we can have two subscriber files:
 *
 *      digest  -  lists users who want to recieve email digest of conf
 *      notify  -  lists users who want instant notification of posts to conf
 *
 * These are secondary, binary files.  The primary source is the 'subs' setting
 * in the user's participation files for the conference.  The subscriber files
 * are automatically updated to to match the participation file settings.
 */


/* Format of records in 'notify' file.  Flags are as defined in subscriber.h.
 * If user is null, this is a blank space in the file.
 */

struct sub_notify
{
    char user[MAX_LOGIN_LEN];		/* User login - not null terminated */
    unsigned char flags;		/* SUB_ flag bits */
};


/* Format of records in 'digest' file.  Flags are as defined in subscriber.h.
 * Period is how often to send a digest.  If X is the number of days since
 * Jan 1, 1970 and (X modulo period) == base or if last is more than period
 * days ago, then today is a day to send a digest.  If user is null, this is
 * a blank space in the file.
 */

struct sub_digest
{
    char user[MAX_LOGIN_LEN];		/* User login - not null terminated */
    unsigned char flags;		/* SUB_ flag bits */
    unsigned char period;		/* Length of period in days */
    unsigned char base;			/* value between 0 and period-1 */
    time_t last;			/* When last digest was sent */
};


/* NOTIFY_FNAME, DIGEST_FNAME
 *   Return name of the desired file
 */

void notify_fname(char *fname, int len)
{
    char *cfd= sdict(VAR_CONFDIR);

    /* Construct file name */
    if (cfd == NULL || cfd[0] == '\0')
	die("Cannot access notify file when no conference is open");
    snprintf(fname,len,"%s/notify",cfd);
}

void digest_fname(char *fname, int len)
{
    char *cfd= sdict(VAR_CONFDIR);

    /* Construct file name */
    if (cfd == NULL || cfd[0] == '\0')
	die("Cannot access digest file when no conference is open");
    snprintf(fname,len,"%s/digest",cfd);
}


/* UPDATE_SUBSCRIBER(<userid>,<string>)
 *   A new value <string> has been stored for the 'subs' setting in a user's
 *   participation file.  Update the subscriber files to reflect that.
 *
 *   The format of the string is
 *       'I <flags>'           - instant
 *       'D <flags> <period>'  - digest
 *   Where <flags> is a decimal integer representing a boolean of SUBS_* flags,
 *   and <period> is an integer representing the period in days.
 */

void update_subscriber(char *user, char *str)
{
    if (bdict(VAR_SUB_INSTANT))
    {
	if (str != NULL && str[0] == 'I')
	{
	    char *f= firstout(str+1," \t");
	    if (!isascii(*f) || !isdigit(*f))
		die("Bad instant subscription string '%s'", str);
	    update_notify(user, atoi(f));
	}
	else
	{
	    update_notify(user, -1);
	}
    }

    if (bdict(VAR_SUB_DIGEST))
    {
	if (str != NULL && str[0] == 'D')
	{
	    char *f= firstout(str+1," \t");
	    char *p= firstout(firstin(f," \t")," \t");
	    int period;
	    if (!isascii(*f) || !isdigit(*f) || !isascii(*p) || !isdigit(*p) ||
	        (period= atoi(p)) < 1)
		die("Bad digest subscription string '%s'", str);
	    /* Base is set to a random value */
	    update_digest(user, atoi(f), period, time(NULL)%period, 0L);
	}
	else
	{
	    update_digest(user, -1, 0, 0, 0L);
	}
    }
}


/* UPDATE_NOTIFY(<userid>,<flags>)
 *   If <flags> is non-negative, insert an entry in the notify file.  If
 *   there is already an entry for the given user, its flags are changed
 *   to the given values, otherwise an entry is added.
 *
 *   If <flags> is negative, delete the user's entry from the notify file,
 *   if there is an entry.
 */

void update_notify(char *user, int flags)
{
    int nfd;
    char fname[BFSZ+1];
    struct sub_notify line;
    off_t offset= 0, blank= -1;

    /* Open file to read/write */
    notify_fname(fname,BFSZ);
    if ((nfd= open(fname,O_CREAT|O_RDWR,0600)) < 0)
    	die("Cannot open notify file %s",fname);

    while (read(nfd, &line, sizeof(struct sub_notify)) ==
	    sizeof(struct sub_notify))
    {
	if (!strncmp(user,line.user,MAX_LOGIN_LEN))
	{
	    /* Found the user */
	    if (flags < 0)
		/* erase him */
		memset(line.user, 0, MAX_LOGIN_LEN);
	    else
		/* set his flags */
		line.flags= flags;

	    /* update the file */
	    lseek(nfd, offset, SEEK_SET);
	    write(nfd, &line, sizeof(struct sub_notify));

	    return;
	}
	else if (blank < 0 && line.user[0] == '\0')
	{
	    /* remember offset of first blank entry we saw */
	    blank= offset;
	}
	offset+= sizeof(struct sub_notify);
    }

    /* User was not found - if we were deleting, we are content */
    if (flags >= 0)
    {
	/* Otherwise, add him to the file, at first blank or at end */
	strncpy(line.user, user, MAX_LOGIN_LEN);
	line.flags= flags;

	if (blank < 0)
	    lseek(nfd, offset, SEEK_SET);
	else
	    lseek(nfd, blank, SEEK_SET);

	write(nfd, &line, sizeof(struct sub_notify));
    }

    close(nfd);
}


/* UPDATE_DIGEST(<userid>,<flags>,<period>,<base>,<time>)
 *   If <flags> is non-negative, insert an entry in the digest file.  If
 *   there is already an entry for the given user, its flags are changed
 *   to the given values, otherwise an entry is added.  The last digest time
 *   is set to the given value only if the user was not previously in the
 *   database.
 *
 *   If <flags> is negative and time is positive, update the user's last
 *   digest time only - do not change flags, period, or base.
 *
 *   If <flags> is negative and time is zero, delete the user's entry from
 *   the digest file, if there is an entry.
 */

void update_digest(char *user, int flags, int period, int base, time_t tm)
{
    int dfd;
    char fname[BFSZ+1];
    struct sub_digest line;
    off_t offset= 0, blank= -1;

    /* Open file to read/write */
    digest_fname(fname,BFSZ);
    if ((dfd= open(fname,O_CREAT|O_RDWR,0600)) < 0)
    	die("Cannot open digest file %s",fname);

    while (read(dfd, &line, sizeof(struct sub_digest)) ==
	    sizeof(struct sub_digest))
    {
	if (!strncmp(user,line.user,MAX_LOGIN_LEN))
	{
	    /* Found the user */
	    if (flags >= 0)
	    {
		/* set his flags */
		line.flags= flags;
		line.period= period;
		line.base= base;
	    }
	    else if (tm == 0)
	    {
		/* erase him */
		memset(line.user, 0, MAX_LOGIN_LEN);
	    }
	    else
	    {
		/* update time only */
		line.last= tm;
	    }

	    /* update the file */
	    lseek(dfd, offset, SEEK_SET);
	    write(dfd, &line, sizeof(struct sub_digest));

	    return;
	}
	else if (blank < 0 && line.user[0] == '\0')
	{
	    /* remember offset of first blank entry we saw */
	    blank= offset;
	}
	offset+= sizeof(struct sub_digest);
    }

    /* User was not found - if we were deleting or updating time,
     * we are content */
    if (flags >= 0)
    {

	/* Otherwise, add him to the file, at first blank or at end */
	strncpy(line.user, user, MAX_LOGIN_LEN);
	line.flags= flags;
	line.period= period;
	line.base= base;
	line.last= tm;

	if (blank < 0)
	    lseek(dfd, offset, SEEK_SET);
	else
	    lseek(dfd, blank, SEEK_SET);

	write(dfd, &line, sizeof(struct sub_digest));
    }

    close(dfd);
}


/* OPEN_NOTIFY()
 * NEXT_NOTIFY(targetflags)
 * CLOSE_NOTIFY()
 *   Routines to walk thought the notify file.  Open_notify() is called to
 *   start a walk.  The walk will include only users who have all the target
 *   flags set.  Next_notify() finds the next matching users.  Returns a
 *   pointer to the user name in a static memory areao of size MAX_LOGIN_LEN
 *   on success.  THE USER NAME IS NOT NECESSARILY NULL TERMINATED!  Returns
 *   NULL on failure.  Close_notify() cleans up after a walk.
 */

static int notify_fd= -1;

void close_notify()
{
    if (notify_fd >= 0) close(notify_fd);
    notify_fd= -1;
}

void open_notify()
{
    char fname[BFSZ+1];

    if (!bdict(VAR_SUB_INSTANT)) return;

    close_notify();

    /* Open file to read/write */
    notify_fname(fname,BFSZ);
    notify_fd= open(fname,O_RDONLY);
}


char *next_notify(int tflags)
{
    static struct sub_notify line;

    if (notify_fd < 0) return NULL;
    while (read(notify_fd, &line, sizeof(struct sub_notify)) ==
	   sizeof(struct sub_notify))
    {
	if (line.user[0] != '\0' && ((line.flags & tflags) == tflags) )
	    return line.user;
    }
    /* End of file */
    close_notify();
    return NULL;
}


/* OPEN_DIGEST(time)
 * NEXT_DIGEST(targetflags)
 * CLOSE_DIGEST()
 *   Routines to walk thought the digest file.  Open_digest() is called to
 *   start a walk.  The walk will include only users who have all the target
 *   flags set, and who are due for a digest on the given time.  Next_digest()
 *   finds the next matching user.  On success, returns a pointer to the user
 *   name in a static memory area of size MAX_LOGIN_LEN.  THE USER NAME IS NOT
 *   NECESSARILY NULL TERMINATED!  On failure returns NULL.  Close_digest()
 *   cleans up after a walk.
 */

static int digest_fd= -1;
static time_t digest_day;

void close_digest()
{
    if (digest_fd >= 0) close(digest_fd);
    digest_fd= -1;
}

void open_digest(time_t at)
{
    char fname[BFSZ+1];

    if (!bdict(VAR_SUB_DIGEST)) return;

    close_digest();

    /* Open file to read/write */
    digest_fname(fname,BFSZ);
    digest_fd= open(fname,O_RDONLY);
    digest_day= at/(24*60*60);
}


char *next_digest(int tflags)
{
    static struct sub_digest line;
    time_t lastday;

    if (digest_fd < 0) return NULL;
    while (read(digest_fd, &line, sizeof(struct sub_digest)) ==
	   sizeof(struct sub_digest))
    {
	if (line.user[0] != '\0' && ((line.flags & tflags) == tflags) )
	{
	    lastday= line.last/(24*60*60);
	    if (lastday < digest_day - line.period ||
		(lastday < digest_day &&
		  ((digest_day % line.period) == line.base) ) )
		return line.user;
	}
    }
    /* End of file */
    close_digest();
    return NULL;
}


/* FUNC_NEXT_SUBSCRIBER:  <target_flags> <type> next_subscriber <uid> 0
 *                        <target_flags> <type> next_subscriber 1
 *
 *   Walk a subscriber list.  Type is (D) or (I) for digest or instant.
 *   Target flags are integers defined as described above.  First call
 *   gives first user.  Next call gives next, until EOF, when we return
 *   1.  Next call after that returns first again.  No way to rewind list.
 */

void func_next_subscriber()
{
    char *type= pop_string();
    int tflags= pop_integer();
    char *user;

    if (type[0] == 'I' || type[0] == 'i')
    {
	if (notify_fd < 0) open_notify();
	user= next_notify(tflags);
    }
    else if (type[0] == 'D' || type[0] == 'd')
    {
	if (digest_fd < 0) open_digest(time(NULL));
	user= next_digest(tflags);
    }
    else
    {
	die("Bad subscriber type: '%s'", type);
    }

    if (user == NULL)
    {
	push_integer(1);
    }
    else
    {
	push_string(strndup(user,MAX_LOGIN_LEN),FALSE);
	push_integer(0);
    }
}
